# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is the Bugzilla Launchpad Plugin.
#
# The Initial Developer of the Original Code is Canonical Ltd.
# Portions created by Canonical are Copyright (C) 2008 Canonical Ltd.
# All Rights Reserved.
#
# Contributor(s): Bradley Baetz <bbaetz@acm.org>
#                 Elliotte Martin <emartin@everythingsolved.com>

package extensions::launchpad::lib::WSLaunchpad;
use strict;
use warnings;

use base qw(Bugzilla::WebService);

import SOAP::Data qw(type);

use Bugzilla::Bug;
use Bugzilla::Constants;
use Bugzilla::Error;
use Bugzilla::Search;
use Bugzilla::Util;
use Bugzilla::WebService::Bugzilla;
use Bugzilla::WebService::Bug;

use Date::Format;
use LWP::UserAgent;
use Time::Zone;

use constant LAUNCHPAD_TIMEOUT => 10; # Seconds before token-check timeout.
use constant LAUNCHPAD_URL => 'https://launchpad.net/';

use constant LOGIN_EXEMPT => {
        login => 1,
};

sub add_comment {
    my $self = shift;
    $self->Bugzilla::WebService::Bug::add_comment(@_);

    my $dbh = Bugzilla->dbh;
    my $comment_id = $dbh->bz_last_key('longdescs', 'comment_id');
    return {'comment_id' => type('int')->value($comment_id) };
}

sub comments {
    my ($self, $params) = @_;

    my $bug_ids = $params->{bug_ids};
    my $comment_ids = $params->{comment_ids};
    (defined $bug_ids || defined $comment_ids)
        || ThrowCodeError('param_required', { param => 'bug_ids or comment_ids' });

    # This is a bit ugly. We can't just get the comments matching
    # 'comment_ids' because the only way to get the comment number
    # is to get all the comments.
    # Also, we need to record all comments from passed in bugs,
    # but only the requested comments from things not passed in,
    # so we need to keep track of the difference.
    # We also need to check for validity up front. So what we do is:
    #  - Get the comments
    #  - Check for validity
    #  - Have a hash that keeps track of whether we want all comments
    #    or not.
    # This interface is sucky!
    # There are possible perf issues, but it *should* be OK for now,
    # given the expected use cases
    $bug_ids = { map { $_ => 1 } @$bug_ids };
    $comment_ids = { map { $_ => 1 } @$comment_ids };
    
    my $dbh = Bugzilla->dbh;
    if (scalar keys %$comment_ids) {
        my $q = 'SELECT DISTINCT bug_id FROM longdescs WHERE comment_id IN ('
                . join(', ', ('?') x scalar(keys %$comment_ids)) . ')';
        my $comment_bug_ids = 
            $dbh->selectcol_arrayref($q, undef, keys %$comment_ids);
        # Set this to 2 so that we can distinguish the two cases
        # (bug wanted vs. comment wanted) later.
        $bug_ids->{$_} ||= 2 foreach (@$comment_bug_ids);
    }

    my $include_fields = $params->{include_fields};
    $include_fields ||= [qw(id number text author time)];
    my %inc = map { $_ => 1 } @$include_fields;

    my $start = $params->{changed_since};
    if (defined $start) {
        # Isn't there a wrapper for this, that the XMLRPC stuff
        # just does automatically???
        $start =~ /^(\d{4})(\d{2})(\d{2})T(\d{2}):(\d{2}):(\d{2})$/
            || ThrowUserError('illegal_date', 
               { date => $start, format => 'YYYYMMDDThh:mm:ss' });
        $start = "$1.$2.$3 $4:$5:$6";
    }

    my %ret;
    foreach my $bug_id (keys %$bug_ids) {
        ValidateBugID($bug_id);

        # Time has to be filtered here, because theres no other way
        # to get the comment count
        my $comments = Bugzilla::Bug::GetComments($bug_id, 'oldest_to_newest');

        my $cnt = -1;

        foreach my $comment (@$comments) {
            $cnt++;

            # Filter to only show wanted comments
            # ie bug wanted or comment wanted
            next unless $bug_ids->{$bug_id} == 1 ||
                $comment_ids->{$comment->{id}};

            # Filter for start time
            next unless $comment->{'time'} ge $start;

            # Filter for comment visibility
            # XXX - This should be filtered somewhere else
            next if $comment->{'isprivate'} && !Bugzilla->user->is_insider;

            my %c;
            $c{id} = type('int')->value($comment->{id}) if $inc{id};
            $c{number} = type('int')->value($cnt) if $inc{number}; 
            $c{text} = type('string')->value($comment->{body}) if $inc{text};
            $c{author} = type('string')->value($comment->{author}->login)
                if $inc{author};
            $c{'time'} = type('dateTime')->value(
                               $self->datetime_format($comment->{'time'}))
                if $inc{'time'};

            push @{$ret{$bug_id}}, \%c;
        }
    }

    return { bugs => \%ret };
}

sub count {
    my ($self, $params) = @_;
    
    my $products = $params->{products};
    my $dbh = Bugzilla->dbh;

    ThrowUserError('no_products') unless scalar(@$products);
    
    foreach my $product_name (@$products){
        ThrowUserError('product_access_denied', {product => $product_name})
            unless Bugzilla->user->can_see_product($product_name);
    }
    
    # Convert product names to their IDs.
    my $product_ids =
        $dbh->selectcol_arrayref('SELECT id FROM products
                                  WHERE ' .
                                  $dbh->sql_in('name', [('?') x @$products]),
                                  undef, @$products);

    my $bug_count =
      $dbh->selectrow_array('SELECT COUNT(bug_id)
                               FROM bugs
                               WHERE '.
                               $dbh->sql_in('product_id', $product_ids));
                            
    return {'count' => type('int')->value($bug_count)};
}

sub get_bugs {
    my ($self, $params) = @_;
    my $ids = $params->{ids};
    my $products = $params->{products};
    (defined $ids || defined $products)
        || ThrowCodeError('param_required', { param => 'ids or products' });

    my $limit = $params->{limit};
    if ($limit) {
        detaint_natural($limit)
            || ThrowUserError('invalid_maxrows', { maxrows => $params->{limit} });
    }
    my $offset = $params->{offset};
    if(defined($offset) && !$limit) {
        ThrowCodeError('param_required', { param => 'limit' });
    }
    elsif (defined $offset) {
        detaint_natural($offset)
            || ThrowUserError('invalid_offset', { offset => $params->{offset} });
    }

    # Use Search.pm so that all the constraints can be applied simply
    # including visibility limits. All parameter validation will happen
    # there.

    # Build up a boolean chart
    # Search takes an object that has a ->param method; CGI suffices.
    my $q = new CGI({});

    # First, bug and product constraints
    # These are a union
    my $chart = 0;
    my $col = 0;
    if (defined $ids) {
        # Need to handle bug ids and aliases
        # We *do* validate these, so that any explicitly
        # listed entries throw an error if they're not viewable
        my @ids;
        foreach my $bug_id (@$ids) {
            ValidateBugID($bug_id);
            push @ids, $bug_id;
        }
        $q->param("field$chart-0-$col", 'bug_id');
        $q->param("type$chart-0-$col", 'anyexact');
        $q->param("value$chart-0-$col", join(',', @ids));
        $col++;
    }
    if (defined $products) {
        # Can't use anyexact in case the product contains a comma
        foreach my $prod (@$products) {
            $q->param("field$chart-0-$col", 'product');
            $q->param("type$chart-0-$col", 'equals');
            $q->param("value$chart-0-$col", $prod);
            $col++;
        }
    }
    $chart++;

    # Now do changed_since
    my $changed_since = $params->{changed_since};
    if (defined $changed_since) {
        # Format incoming iso8601 dateTime correctly
        # This will error out if the incoming value isn't parseable
        # XXX - shouldn't Bugzilla::Search be doing this for date
        # searches?
        $changed_since = Bugzilla::Search::SqlifyDate($changed_since);

        # No greaterthanorequal, so do NOT lessthan
        $q->param("negate$chart", 1);
        $q->param("field$chart-0-0", 'delta_ts');
        $q->param("type$chart-0-0", 'lessthan');
        $q->param("value$chart-0-0", $changed_since);
        $chart++;
    }

    my $s = new Bugzilla::Search('fields' => ['bugs.bug_id'],
                                 'params' => $q,
                                 'order' => ['bugs.bug_id']);

    my $dbh = Bugzilla->dbh;

    my $sql = $s->getSQL();
    $sql .= ' ' . $dbh->sql_limit($limit, $offset) if ($limit);

    my $bugs = $dbh->selectcol_arrayref($sql);

    my @return;
    foreach my $bug_id (@$bugs) {
        # Note - checking if the user can see the bug is handled by Search.pm
        my $bug = new Bugzilla::Bug($bug_id);

        # Timetracking fields are deleted if the user doesn't belong to
        # the corresponding group.
        unless (Bugzilla->user->in_group(Bugzilla->params->{'timetrackinggroup'})) {
            delete $bug->{'estimated_time'};
            delete $bug->{'remaining_time'};
            delete $bug->{'deadline'};
        }
        # This is done in this fashion in order to produce a stable API.
        # The internals of Bugzilla::Bug are not stable enough to just
        # return them directly.
        my $creation_ts = $self->datetime_format($bug->creation_ts);
        my $delta_ts    = $self->datetime_format($bug->delta_ts);
        my $status      = $bug->status;
        my %item;
        $item{'id'}               = type('int')->value($bug->bug_id);

        $item{'assigned_to'}      = type('string')->value($bug->assigned_to->login);
        $item{'component'}        = type('string')->value($bug->component);
        $item{'creation_time'}    = type('dateTime')->value($creation_ts);
        $item{'dupe_of'}          = type('int')->value($bug->dup_id)
            if $bug->dup_id;
        $item{'is_open'}          = type('boolean')->value($status->is_open);
        $item{'last_change_time'} = type('dateTime')->value($delta_ts);
        $item{'priority'}         = type('string')->value($bug->priority);
        $item{'product'}          = type('string')->value($bug->product);
        $item{'resolution'}       = type('string')->value($bug->resolution);
        $item{'severity'}         = type('string')->value($bug->bug_severity);
        $item{'status'}           = type('string')->value($status->name);
        $item{'summary'}          = type('string')->value($bug->short_desc);

        # Let people poke inside if they know what they're doing...
        $item{'internals'}        = $bug;

        if (Bugzilla->params->{'usebugaliases'}) {
            $item{'alias'} = type('string')->value($bug->alias);
        }
        else {
            # For API reasons, we always want the value to appear, we just
            # don't want it to have a value if aliases are turned off.
            $item{'alias'} = undef;
        }

        push(@return, \%item);
    }

    return { bugs => \@return };
}

sub set_link {
    my ($self, $params) = @_;
    
    my $bug_id = $params->{id};
    my $launchpad_id = $params->{launchpad_id};

    #The user must login in order add a launchpad_id
    Bugzilla->login(LOGIN_REQUIRED);

    # Check parameters
    defined $bug_id  || ThrowCodeError('missing_bug_id');

    ValidateBugID($bug_id);

    my $bug = new Bugzilla::Bug($bug_id);

    Bugzilla->user->can_edit_product($bug->product_id)
        || ThrowUserError("product_edit_denied", {product => $bug->product});

    my $dbh = Bugzilla->dbh;
    
    my $old_launchpad_id = $dbh->selectrow_array(
            "SELECT launchpad_id
               FROM bugs WHERE bug_id = ?", undef, $bug_id);

    my $privs;
    $bug->check_can_change_field('launchpad_id', 
                                 $old_launchpad_id, $launchpad_id, \$privs)
            || ThrowUserError('illegal_change', 
                              { field => 'launchpad_id', privs => $privs });

    if (defined ($launchpad_id)) {
        if (!defined ($old_launchpad_id) ||
             $launchpad_id != $old_launchpad_id)
        {
               $dbh->do("UPDATE bugs SET launchpad_id = ?
                         WHERE bug_id = ?", undef, $launchpad_id, $bug_id);
        }
    }
    elsif (defined ($old_launchpad_id)){
            $dbh->do("UPDATE bugs SET launchpad_id = NULL
                      WHERE bug_id = ?", undef, $bug_id);
    }

    return {'launchpad_id' => type('int')->value($old_launchpad_id)};
}

sub login {
    my ($self, $params) = @_;

    my $token = $params->{token};
    $token || ThrowCodeError('param_required', { param => 'token' });

    # Taken from Bugzilla::Update
    my $ua = LWP::UserAgent->new();
    $ua->timeout(LAUNCHPAD_TIMEOUT);
    $ua->protocols_allowed(['https']);
    # If the URL of the proxy is given, use it, else get this information
    # from the environment variable.
    my $proxy_url = Bugzilla->params->{'proxy_url'};
    if ($proxy_url) {
        $ua->proxy(['http', 'https'], $proxy_url);
    } else {
        $ua->env_proxy;
    }

    my $resp = $ua->post(LAUNCHPAD_URL . 'token/' . url_quote($token)
                         . '/+bugtracker-handshake');
    
    if ($resp->code() != 200) {
        ThrowUserError('launchpad_invalid_token', {
            token => $token,
            code => $resp->code,
            content => $resp->content,
        });
    }

    my $auth = Bugzilla::Auth->new({Login => 'Launchpad'});
    my $user = $auth->login(LOGIN_REQUIRED);

    return { user_id => type('int')->value($user->id) };
}

sub plugin_version(){	
    my $plugins = Bugzilla::WebService::Bugzilla::extensions();
    return {'version' => $plugins->{'extensions'}->{'launchpad'}->{'version'}};
}

sub time {
    my ($self) = @_;
    my $dbh = Bugzilla->dbh;

    my $db_time = $dbh->selectrow_array('SELECT LOCALTIMESTAMP(0)');
    return {
        local_time => type('dateTime')->value($self->datetime_format($db_time)),
        utc_time   => type('dateTime')->value(
                          time2str('%Y%m%dT%H:%M:%S', time, 'UTC')),
        tz_name => type('string')->value(uc tz_name()),
    };
}

1;

__END__

=head1 NAME

extensions::launchpad::lib::WSLaunchpad - Webservice functions for the
Launchpad extention

=head1 DESCRIPTION

This package contains various webservices calls for the Launchpad plugin

=head1 CALLS

There are a number of XML-RPC calls.

=over

=item C<add_comment> B<EXPERIMENTAL>

The same as L<Bugzilla::WebService::Bug/add_comment>, except it returns
a single item C<comment_id>, which is an C<int> that represents the ID
of the added comment.

=item C<comments> B<UNSTABLE>

=over

=item B<Description>

This allows you to get comments for a set of bugs.

=item B<Params>

=over

=item C<bug_ids> (array of ints) - An array of bugs IDs to return
comments for. At least one of C<bug_ids> or C<comment_ids>
is B<required>.

=item C<comment_ids> (array of ints) - An array of integer comment_ids 
to return. At least one of C<bug_ids> or C<comment_ids>
is B<required>.

=item C<include_fields> (array of strings) - An array of field names to be
included in the returned hash. If this parameter is not present, all fields
are returned. If it is present, then only those fields will be returned. For
forwards/backwards compatibility, it is not an error to specify additional
fields.

=item C<changed_since> (dateTime) - Only return comments that are newer
than the specified date.

=back

=item B<Returns>

A single item, C<bugs>, which is a hash, where the keys are C<bug_id>s and the
values are an array of comments. Each comment is a hash, containing:

=over

=item id

C<int> The globally unique ID for the comment.

=item number

C<int> The comment number for the comment. (You could link to this comment
using C<#c8>, for example, if this was comment number 8 on the bug).

=item text

C<string> The actual text of the comment.

=item author

C<string> The login name of the comment's author.

=item time

C<dateTime> The time (in Bugzilla's timezone) that the comment was added.

=back

=item B<Errors>

Same as L</get_bugs>.

=back


=item C<count> B<UNSTABLE>

=over

=item B<Description>

Returns the bug count for a list of products.

=item B<Params>

=over

=item C<products> (array) B<Required> - An array of product ID's.

=back

=item B<Returns>

A hash with one element, C<count>. This is the bug count for the list of products.

=item B<Errors>

=over

=item 106 (Invalid Product)

Either you didn't specify a product, one or more products don't exist, or
you don't have permission to access bugs for one or more of the products.

=back

=back


=item C<get_bugs> B<UNSTABLE>

=over

=item B<Description>

Gets information about particular bugs in the database.

=item B<Params>

=over

=item C<ids>

An array of numbers and strings.

If an element in the array is entirely numeric, it represents a bug_id
from the Bugzilla database to fetch. If it contains any non-numeric 
characters, it is considered to be a bug alias instead, and the bug with 
that alias will be loaded. 

Note that it's possible for aliases to be disabled in Bugzilla, in which
case you will be told that you have specified an invalid bug_id if you
try to specify an alias. (It will be error 100.)

=item C<products>

An array of product names. All bugs in the specified products are returned.

If both C<ids> and C<products> are present, the union of the two are returned.

=item C<changed_since>

A C<dateTime> field. Only bugs changed on or after this time are returned.
Newly-filed bugs are considered "changed" at the time they were filed, and
will be included in the returned results.

=item C<limit>

Return at most this many results.

=item C<offset>

Starts returning bugs from this point in the list. Requires C<limit> to be
set.

Note that while using C<limit> and C<offset> to handle a large list in
chunks is possible, it is impossible to guarantee that no changes are
being made between runs (eg changing the product of a bug), so particular
bugs might be skipped or returned multiple times.

=back

=item B<Returns>

A hash containing a single element, C<bugs>. This is an array of hashes. 
Each hash contains the following items:

=over

=item id

C<int> The numeric bug_id of this bug.

=item alias

C<string> The alias of this bug. If there is no alias or aliases are 
disabled in this Bugzilla, this will be an empty string.

=item C<assigned_to>

C<string> The login name of the current assignee.

=item C<component>

C<string> The name of the component the bug is in.

=item creation_time

C<dateTime> When the bug was created.

=item C<dupe_of>

C<int> If this bug is marked as a duplicate of another, this field will
exist and contain the number of the bug that this bug is a duplicate of.

If this bug is not a duplicate, this field will be empty.

=item C<is_open>

C<boolean> Whether or not the bug is in a status that is considered to
be "open".

=item last_change_time

C<dateTime> When the bug was last changed.

=item C<priority>

C<string> The priority set on this bug.

=item C<product>

C<string> The name of the product this bug is in.

=item C<resolution>

C<string> If this bug is in a closed state, this is the name of the
resolution it was given.

=item C<severity>

C<string> The severity set on this bug.

C<status>

C<string> The current status (state) this bug is in.

=item summary

C<string> The summary of this bug.

=item internals B<UNSTABLE>

A hash. The internals of a L<Bugzilla::Bug> object. This is extremely
unstable, and you should only rely on this if you absolutely have to. The
structure of the hash may even change between point releases of Bugzilla.

=back

=item B<Errors>

=over

=item 100 (Invalid Bug Alias)

If you specified an alias and either: (a) the Bugzilla you're querying
doesn't support aliases or (b) there is no bug with that alias.

=item 101 (Invalid Bug ID)

One of the bug ids you specified doesn't exist in the database.

=item 102 (Access Denied)

You do not have access to a bug id you specified.

=item 10001 (Invalid Offset)

You specified an C<offset> that wasn't numeric.

=back

=back


=item C<login> B<EXPERIMENTAL>

=over

=item B<Description>

Allows launchpad to log in. The launchpad server URL and Bugzilla username
must be configured in the preferences.

This routine takes a token provided by the remote server. Bugzilla does a
POST to https://launchpad.net/token/C<token>/+bugtracker-handshake using 
the provided token.

If that POST returns a 200 response, then the token is accepted, and Bugzilla
returns the standard authentication cookie headers. Otherwise, the request
fails.

The returned headers are then valid in any further request to serve as
authentication. Note that the Bugzilla server tends to expire these logins
after 30 days, but they could expire earlier than that if the admin manually
logs out all users. (It is impossible to know whether or not your cookie
is still valid without using it.)

=item B<Params>

=over

=item C<token> (string) B<Required> - The token to use.

=back

=item B<Returns>

=over

=item C<user_id> (integer) - The user_id of the configured launchpad user.

=back

=item B<Errors>

=over

=item 10300 (Invalid Token)

The token you provided was not valid (the login failed).

=item -10301 (Plugin Not Configured)

The user hasn't run F<checksetup.pl> since installing the Launchpad
plugin.

=back

=back

=item C<set_link> B<EXPERIMENTAL>

=over

=item B<Description>

Takes a Launchpad bug ID and a Bugzilla Bug ID and sets a link between them.
If the Launchpad bug ID is not specified, any existing link will be deleted.

=item B<Params>

=over

=item C<id> (int) B<Required> - The id or alias of the Bugzilla bug.

=item C<launchpad_id> (int) - The Launchpad bug ID.

=back

=item B<Returns>

C<launchpad_id> (int) The previously-existing Launchpad ID, if there was one.

=item B<Errors>

=over

=item 100 (Invalid Bug Alias)

If you specified an alias and either: (a) the Bugzilla you're querying
doesn't support aliases or (b) there is no bug with that alias.

=item 101 (Invalid Bug ID)

The id you specified doesn't exist in the database.

=item 108 (Bug Edit Denied)

You did not have the necessary rights to edit the bug.

=back

=back

=item C<time> B<EXPERIMENTAL>

Returns three elements:

=over

=item local_time (dateTime)

The current time in the local timezone, according to the Bugzilla
database server.

=item utc_time (dateTime)

The current time in UTC, according to the Bugzilla web server.

=item tz_name (string)

Current time zone

=back

=item C<plugin_version> B<EXPERIMENTAL>

=over

=item B<Description>

Returns the Launchpad plug-in version number.

=item B<Returns>

A hash with a single item, C<version>, that is the version as a
string.

=back

=back
